function [MLE] = MLE_hierarchical_HMM_model(y,kbar_mixture,kbar,With_leverage,Simu_prev)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Realized Variances: Function that estimates the FHMV model with/out leverage effect.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% y : Time series data  (dimension: Tx1 or if with leverage Tx2)
%%%%% kbar_mixture : Number of components of the mixture block.
%%%%% kbar : Number of components of the persistent block.
%%%%% With_leverage : if =1 then leverage effect (default = 0).
%%%%% Simu_prev : Structure with the starting point of the optimization and the number of kbar (optional)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% To run : [Simu] = max_hierarchical_HMM_model(y,10,10,1);
%%%% FHMV model with 10 mixture components, kbar=10 (5 secondary
%%%% components) and with leverage effect.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if(nargin<5)
    Simu_prev = 0;
    if(nargin<4)
       With_leverage = 0;
    end
end

%%%%%%% Additional parameters
nb_free_MS = 0; % nb_free_MS : Number of Component C_t^(i) that is free of parametrization (Possible values : 1,2,3,4 and 6 if kbar_mixture = 2)
distribution_choice = 'Gamma'; % distribution_choice : Distribution of the innovation: Gamma, Inverse Gamma or Log-logistic.
model_variance = 1; % Model defined for variance, if set to zero: defined for standard errors.

switch distribution_choice
    case 'Log-Logistic'
        gamma_dist = 2;
    case 'IGamma'
        gamma_dist = 0;
    otherwise
        distribution_choice = 'Gamma';
        gamma_dist = 1;
end

const_P = 2+nb_free_MS;
if(With_leverage~=0)
   nb_series = min(size(y));
   if(nb_series==1)
       warning('****** Modeling the leverage effect can only be carried out with an additional series of returns');
       disp('****** The model does not take into account the leverage effect ********');
       With_leverage = 0;
   elseif(sum(y(:,2)<0)==0)
       disp('****** The additional series does not have negative values --> no leverage effect ********');
       With_leverage = 0;
   end
end

if(isstruct(Simu_prev))
    if(isfield(Simu_prev,'theta'))
        init_theta = Simu_prev.theta;
        size_init_theta = max(size(init_theta));
        kbar_prev = Simu_prev.kbar;
    else
        size_init_theta = 0;
        kbar_prev = 0;
    end
else
    size_init_theta = 0;
    kbar_prev = 0;
end


regime = 2^(kbar);
taille = max(size(y));


if(const_P<4 && const_P>0)
    nb_param_P = const_P;
else
   nb_param_P = kbar_mixture+kbar; 
   warning('const P should be between 1 and 6');
end

nb_param_lev = 0;
if(With_leverage~=0)
    nb_param_lev = 2;
end

nb_param_sigma = 1+1+2+2+nb_free_MS;
nb_param_tot = nb_param_P+nb_param_sigma+nb_param_lev;
nb = 1;  %%% Number of Maximisations with a new starting point.
if(size_init_theta==nb_param_tot)
    nb = 1;
end


disp('*******************************************************')
disp(['Estimation of the hierarchical MS model with ' num2str(kbar_mixture) ' Mixture components and ' num2str(kbar) ' Multifractal components'])
disp(['Nb parameters for the Mean : ' num2str(nb_param_sigma-1)])
disp(['Nb parameters for the transition matrix : ' num2str(nb_param_P)])
disp(['Nb parameters for the leverage effet : ' num2str(nb_param_lev)])
disp(['Distribution of the innovation : ' distribution_choice])
disp('*******************************************************')
disp(['Total parameters : ' num2str(nb_param_tot)])
disp('*******************************************************')

nb_iter_dep = 50;
nb_chain = ceil(nb_param_tot*1.5);
param_chain = zeros(nb_param_tot,nb_chain);
likelihood_chain = zeros(nb_chain,1);
outlier = ceil(0.2*nb_chain);
nb_chg = ceil(0.1*nb_iter_dep);
time = tic;
for i=1:nb
    test = 0;
    count = 0;
    disp('*******************************************************')
    disp('Searching for a good starting point... ')
    disp('*******************************************************')    
    if(i==1 && size_init_theta==nb_param_tot)
        theta_current = init_theta;
        log_dens = -1*launch_FB_sigma_MLE_for_hierarchical_HMM(y,taille,regime,kbar_mixture,theta_current,nb_param_sigma,With_leverage,gamma_dist,model_variance);
        if(log_dens<1e9)
            test = 1;
        end
    end
    if(test==0)
        disp('Looking for a starting point... First phase out of three...');
        for q=1:nb_chain
            for deb=1:10
                test = 0;
                while(test==0)
                    
                    theta_current = zeros(nb_param_tot,1);
                    theta_current(1) = log(var(y(:,1))) + randn();
                    
                    log_unc_var = log(mean(y(:,1)));
                    theta_current(2) = randn() + log_unc_var;
                    theta_current(3:nb_param_sigma-2) = rand(2+nb_free_MS,1)*0.8;
                    theta_current(3:nb_param_sigma-2) = log(theta_current(3:nb_param_sigma-2));%./(1-theta_current(3:nb_param_sigma-2)));
                    theta_current(nb_param_sigma-1:nb_param_sigma) = rand(2,1)*0.8;
                    theta_current(nb_param_sigma-1:nb_param_sigma) = log(theta_current(nb_param_sigma-1:nb_param_sigma)./(1-theta_current(nb_param_sigma-1:nb_param_sigma)));
                    
                    
                    for z=1:nb_param_P
                        theta_current(nb_param_sigma+z) = 1.5+randn();
                    end
                    if(nb_param_lev~=0)
                        if(nb_param_lev==2)
                            theta_current(nb_param_sigma+nb_param_P+1) = 1.5+randn();
                            theta_current(nb_param_sigma+nb_param_P+2) = -2+randn();
                        else
                            for z=1:nb_param_lev
                                theta_current(nb_param_sigma+nb_param_P+z) = 1+randn();
                            end
                        end
                    end
                    
                    count = count +1;
                    log_dens = -1*launch_FB_sigma_MLE_for_hierarchical_HMM(y,taille,regime,kbar_mixture,theta_current,nb_param_sigma,With_leverage,gamma_dist,model_variance);
                    if(log_dens<1e9)
                        test = 1;
                    end
                    
                    
                end
                if(deb==1)
                    param_chain(:,q) = theta_current;
                    likelihood_chain(q) = log_dens;
                elseif(log_dens<likelihood_chain(q))
                    likelihood_chain(q) = log_dens;
                    param_chain(:,q) = theta_current;                    
                end
            end
        end
        disp('Looking for a starting point... Second phase...');
        likelihood_chain1 = likelihood_chain;
        for q=1:nb_chain
            theta_current = param_chain(:,q);
            option_unc = optimset('MaxIter',10,'MaxFunEvals',100,'FinDiffType','central');
            [theta_current,log_dens] = fminunc(@(theta)  -1*launch_FB_sigma_MLE_for_hierarchical_HMM(y,taille,regime,kbar_mixture,theta,nb_param_sigma,With_leverage,gamma_dist,model_variance),theta_current,option_unc);
            likelihood_chain(q) = log_dens;
            param_chain(:,q) = theta_current;
        end
        likelihood_chain2 = likelihood_chain;
        disp('Looking for a starting point... Third phase...');
        disp(['In process : ' num2str(0) ' %'])
        for k=1:nb_iter_dep
            
            for q=1:nb_chain
                [theta_prop] = DREAM(param_chain,q,0,nb_chain,k);
                log_dens = -1*launch_FB_sigma_MLE_for_hierarchical_HMM(y,taille,regime,kbar_mixture,theta_prop,nb_param_sigma,With_leverage,gamma_dist,model_variance);
                if(log_dens<likelihood_chain(q))
                    likelihood_chain(q) = log_dens;
                    param_chain(:,q) = theta_prop;
                end
            end
            
            if(mod(k,nb_chg)==0 && k<nb_iter_dep)
                disp(['In process : ' num2str(100*k/nb_iter_dep) ' %'])
                [~, indice] = sort(-1*likelihood_chain);
                for z=1:outlier
                    theta_current = param_chain(:,indice(end-outlier+z));
                    option_unc = optimset('MaxIter',10,'MaxFunEvals',500,'FinDiffType','central');
                    [theta_current,log_dens] = fminunc(@(theta) -1*launch_FB_sigma_MLE_for_hierarchical_HMM(y,taille,regime,kbar_mixture,theta,nb_param_sigma,With_leverage,gamma_dist,model_variance),theta_current,option_unc);
                   likelihood_chain(indice(z)) = log_dens;
                   param_chain(:,indice(z)) = theta_current;
                   
                end
            end
        end
        disp([likelihood_chain likelihood_chain2 likelihood_chain1])     
        [log_dens,ind] = min(likelihood_chain);
        theta_current = param_chain(:,ind);
    end
    disp('*******************************************************')
    disp(['Density of the starting point : ' num2str(-1*log_dens)])
    disp('*******************************************************')
    val_theta = theta_current;
    if(kbar_prev==kbar)
        option_search = optimset('MaxIter',30,'MaxFunEvals',30,'Display','iter');
        [val_theta,log_dens] = fminsearch(@(theta) -1*launch_FB_sigma_MLE_for_hierarchical_HMM(y,taille,regime,kbar_mixture,theta,nb_param_sigma,With_leverage,gamma_dist,model_variance),val_theta,option_search);
    else
        option_unc = optimset('MaxIter',1e7,'MaxFunEvals',1e7,'Display','iter','FinDiffType','central');
        [val_theta] = fminunc(@(theta) -1*launch_FB_sigma_MLE_for_hierarchical_HMM(y,taille,regime,kbar_mixture,theta,nb_param_sigma,With_leverage,gamma_dist,model_variance),val_theta,option_unc);
        option_search = optimset('MaxIter',10000,'MaxFunEvals',10000,'Display','iter');
        [val_theta,log_dens] = fminsearch(@(theta) -1*launch_FB_sigma_MLE_for_hierarchical_HMM(y,taille,regime,kbar_mixture,theta,nb_param_sigma,With_leverage,gamma_dist,model_variance),val_theta,option_search);
    end

    log_dens = -1*log_dens; 
    if(i==1)
        max_dens = log_dens;
        max_theta = val_theta;
        
    elseif(max_dens<log_dens)
       max_dens = log_dens;
       max_theta = val_theta;
    end
    
end
disp('*******************************************************')
disp('Computing output variables...')
disp('*******************************************************')

MLE.time_in_min = toc(time)/60;
MLE.theta = max_theta;
MLE.dens = max_dens;
nb_iter_back = 200;
time_like = tic;
[~,sn_sig,sigma,P,forward,theta_lev,var_t_lev,val_theta_sigma,theta_P,val_lev_t_component,gamma_param,fct_leverage,P_kbar,P_mixture,sigma_mixture] = launch_FB_sigma_MLE_for_hierarchical_HMM(y,taille,regime,kbar_mixture,max_theta,nb_param_sigma,With_leverage,gamma_dist,model_variance,nb_iter_back);
MLE.time_to_eval_likelihood = toc(time_like);


MLE.var_t_lev = var_t_lev;
MLE.fct_leverage = fct_leverage;
MLE.lev_t_component = val_lev_t_component;
MLE.theta_lev = theta_lev';
var_t = zeros(taille,nb_iter_back);
for i=1:nb_iter_back
   var_t(:,i) =  sigma(sn_sig(:,i));
end
MLE.var_t = var_t;
MLE.sigma = sigma;
MLE.val_theta_sigma = val_theta_sigma;
MLE.theta_P = theta_P';
MLE.P = P;
MLE.gamma_param = gamma_param;
MLE.kbar_mixture = kbar_mixture;
MLE.kbar = kbar;
MLE.const_P = [num2str(nb_param_P)  ' param for P'];
if(With_leverage~=0)
    MLE.levera = 'With leverage';
    MLE.kbar_leverage = min(size(val_lev_t_component));
else
    MLE.levera = 'Without leverage';
    MLE.kbar_leverage = 0;
end
MLE.distribution = distribution_choice;
if(gamma_dist==1)
    MLE.distribution = 'Gamma';
    MLE.dist_gamma_param = [1/exp(MLE.theta(1)) exp(MLE.theta(1))];
elseif(gamma_dist==0)
    MLE.distribution = 'IGamma';
    MLE.dist_gamma_param = [(exp(MLE.theta(1))+1) exp(MLE.theta(1))];
else
    MLE.distribution = 'Log-Logistic';
    MLE.dist_gamma_param = [(exp(MLE.theta(1))+1) exp(MLE.theta(1))]; 
end

MLE.forward = forward;
nb_reg =  max(size(sigma));

%%%%%%%%%%%%%%%%%%%%%%% Retrieving Persistent and mixture Components
MLE.etat_sig = sn_sig;
kbar_tot = kbar;
prob_over_t_forward = zeros(taille,kbar_tot);
prob_over_t_backward = zeros(taille,kbar_tot);
[vect] = create_indicator_for_states(kbar_tot);

backward = zeros(size(forward));
for j=1:nb_reg
    backward(:,j) = sum(sn_sig==j,2);
end
backward = backward/nb_iter_back;

for j=1:kbar_tot
   ind = vect(j,:)==1;
   prob_over_t_forward(:,j) = sum(forward(:,ind),2);
   prob_over_t_backward(:,j) = sum(backward(:,ind),2);
end


MLE.prob_over_t_forward = prob_over_t_forward;
MLE.prob_over_t_backward = prob_over_t_backward;
MLE.y = y;
MLE.backward = backward;

MLE.P_kbar = P_kbar;
MLE.P_mixture = P_mixture;
MLE.sigma_mixture =  sigma_mixture;
MLE.f_mixture = val_theta_sigma(2:kbar_mixture);
MLE.f_MS = val_theta_sigma(kbar_mixture+1:end-2);
MLE.model_variance = model_variance;



%%%%%%%%%%%%%%%%%%%%%%% Prediction
nb_forecast = 100;
[MLE] = prediction_hierarchical_HMM(MLE,nb_forecast);

%%%%%%%%%%%%%%%%%%%%%%% Simulation of the process given the MLE.
[MLE.DGP] = generate_process(MLE,taille);
